@@ -30,17 +30,23 @@ class MetadataNames:
     GROUP = 'USER_GROUP'
     PINBOARD = 'PINBOARD_ANSWER_BOOK'
     LIVEBOARD = 'PINBOARD_ANSWER_BOOK'
-    WORKSHEEET = 'LOGICAL_TABLE'
+    WORKSHEET = 'LOGICAL_TABLE'
     CONNECTION = 'DATA_SOURCE'
     ANSWER = 'QUESTION_ANSWER_BOOK'
     TABLE = 'LOGICAL_TABLE'
     TAG = 'TAG'
+    VIEW = 'LOGICAL_TABLE'
+    COLUMN = 'LOGICAL_COLUMN'
+    JOIN = 'LOGICAL_RELATIONSHIP'
+    DATA_SOURCE = 'DATA_SOURCE'
 
 
 class MetadataSubtypes:
     WORKSHEET = 'WORKSHEET'
     TABLE = 'ONE_TO_ONE_LOGICAL'
-    VIEW = 'USER_DEFINED'
+    VIEW = 'AGGR_WORKSHEET'
+    USER_UPLOAD = 'USER_DEFINED'
+    PRIVATE_WORKSHEET = 'PRIVATE_WORKSHEET'
 
 
 class MetadataSorts:
@@ -134,6 +140,8 @@ class TSRestApiV1:
         # TS documentation shows the /tspublic/v1/ portion but it is always preceded by {server}/callosum/v1/
         self.base_url = '{server}/callosum/v1/tspublic/v1/'.format(server=self.server)
         self.non_public_base_url = '{server}/callosum/v1/'.format(server=self.server)
+        # V2 REST API for basic implementation
+        self.v2_base_url = '{server}/tspublic/rest/v2/'.format(server=self.server)
 
     #
     # Session management calls
@@ -160,6 +168,55 @@ class TSRestApiV1:
         response.raise_for_status()
         return True
 
+    #
+    # V2 Session management
+    #
+
+    def session_login_v2(self,  username: Optional[str] = None, password: Optional[str] = None,
+                         token: Optional[str] = None) -> bool:
+        endpoint = 'session/login'
+
+        url = self.v2_base_url + endpoint
+
+        if token is not None:
+            response = self.session.post(url=url, headers={"Authorization": "Bearer {}".format(token)})
+        elif username is not None and password is not None:
+            json_post_data = {'userName': username, 'password': password, 'rememberMe': 'true'}
+            response = self.session.post(url=url, json=json_post_data)
+        else:
+            raise Exception("If using username/password, must include both")
+
+        # HTTP 204 - success, no content
+        response.raise_for_status()
+        return True
+
+    # This is only the V2 API bearer token, not the Trusted Authentication token from V1
+    def get_token_v2(self,  username: Optional[str] = None, password: Optional[str] = None,
+                     secret_key: Optional[str] = None, token_expiry_duration: int = 300,
+                     access_level: str = "FULL", ts_object_id: Optional[str] = None) -> Dict:
+        endpoint = 'session/gettoken'
+
+        url = self.v2_base_url + endpoint
+
+        json_post_data = {'accessLevel': access_level,
+                          'tokenExpiryDuration': token_expiry_duration}
+
+        if secret_key is not None:
+            json_post_data['secretKey'] = secret_key
+        elif username is not None and password is not None:
+            json_post_data['userName'] = username
+            json_post_data['password'] = password
+        else:
+            raise Exception("If using username/password, must include both")
+
+        if ts_object_id is not None:
+            json_post_data['tsObjectId'] = ts_object_id
+
+        response = self.session.post(url=url, json=json_post_data, headers={'Accept-Language': 'en_US'})
+
+        response.raise_for_status()
+        return response.json()
+
     #
     # Root level API methods found below, divided into logical separations
     #
@@ -505,7 +562,7 @@ class TSRestApiV1:
 
         post_data = {
             'name': group_name,
-            'displayname': display_name,
+            'display_name': display_name,
             'grouptype': group_type,
             'visibility': visibility
         }
@@ -679,9 +736,16 @@ class TSRestApiV1:
                                    sort: str = 'DEFAULT', sort_ascending: bool = True,
                                    filter: Optional[str] = None, fetchids: Optional[List[str]] = None,
                                    skipids: Optional[List[str]] = None, tagname: Optional[List[str]] = None,
-                                   category: Optional[str] = None, batchsize: int = -1, offset: int = -1) -> Dict:
+                                   category: Optional[str] = None, batchsize: int = -1, offset: int = -1,
+                                   auto_created: Optional[bool] = None) -> Dict:
         endpoint = 'metadata/listobjectheaders'
 
+        # Tables, Worksheets and Views all have the same object_type, with a sub-type to vary
+        # This code allows sending the Sub-Type into Object_type and still run correctly
+        if object_type in [MetadataSubtypes.TABLE, MetadataSubtypes.VIEW, MetadataSubtypes.WORKSHEET]:
+            subtypes = [object_type]
+            object_type = MetadataNames.TABLE
+
         url_params = {
             'type': object_type,
             'sort': sort.upper(),
@@ -702,6 +766,8 @@ class TSRestApiV1:
             url_params['batchsize'] = batchsize
         if category is not None:
             url_params['category'] = category
+        if auto_created is not None:
+            url_params['auto_created'] = str(auto_created).lower()
 
         url = self.base_url + endpoint
         response = self.session.get(url=url, params=url_params)
@@ -746,6 +812,12 @@ class TSRestApiV1:
                       author_guid: Optional[str] = None):
         endpoint = 'metadata/list'
 
+        # Tables, Worksheets and Views all have the same object_type, with a sub-type to vary
+        # This code allows sending the Sub-Type into Object_type and still run correctly
+        if object_type in [MetadataSubtypes.TABLE, MetadataSubtypes.VIEW, MetadataSubtypes.WORKSHEET]:
+            subtypes = [object_type]
+            object_type = MetadataNames.TABLE
+
         url_params = {
             'type': object_type,
             'sort': sort.upper(),
@@ -821,25 +893,23 @@ class TSRestApiV1:
         else:
             j = response.json()
             # JSON error response checking
+
+            # It is possible in a multiple file upload that some validated and others have errors
+
             if 'object' in j:
                 for k in j['object']:
                     if 'info' in k:
                         # Older versions wrapped the errors in 'info'
                         if k['info']['status']['status_code'] == 'ERROR':
                             # print(k['info']['status']['error_message'])
-                            raise SyntaxError(k['info']['status']['error_message'])
-                        else:
-                            return response.json()
+                            raise SyntaxError(j['object'])
                     # Recent versions return as 'response'
                     elif 'response' in k:
                         if k['response']['status']['status_code'] == 'ERROR':
                             # print(k['info']['status']['error_message'])
-                            raise SyntaxError(k['response']['status']['error_message'])
-                        else:
-                            return response.json()
-                    else:
-                        return response.json()
-
+                            raise SyntaxError(j['object'])
+                # If no errors are raised, just return the regular response
+                return response.json()
             else:
                 return response.json()
 
@@ -876,6 +946,40 @@ class TSRestApiV1:
                 raise Exception()
         return tml_obj
 
+    # Method to retrieve the all of the associate objects and retrieve original object and a dict of name:guid mapping
+    def metadata_tml_export_with_associations_map(self, guid: str) -> (OrderedDict, Dict):
+        # Always returns a Python Dict, converted from a request to the API to receive in JSON
+        endpoint = 'metadata/tml/export'
+
+        post_data = {
+            'export_ids': json.dumps([guid]),
+            'formattype': 'JSON',
+            'export_associated': 'true'
+        }
+
+        url = self.base_url + endpoint
+        # TML import is distinguished by having an {'Accept': 'text/plain'} header on the POST
+        response = self.session.post(url=url, data=post_data, headers={'Accept': 'text/plain'})
+        response.raise_for_status()
+        # Extra parsing of some 'error responses' that come through in JSON response on HTTP 200
+        self.raise_tml_errors(response=response)
+
+        # TML API returns a JSON response, with the TML document
+        # object_pairs_hook forces an OrderedDict
+        tml_json_response = response.json(object_pairs_hook=OrderedDict)
+        objs = tml_json_response['object']
+
+        # The first object will be the requested object
+        tml_str = objs[0]['edoc']
+        tml_obj = json.loads(tml_str)
+
+        name_guid_map = {}
+
+        for obj in objs:
+            name_guid_map[obj['info']['name']] = obj['info']['id']
+
+        return tml_obj, name_guid_map
+
     def metadata_tml_export_string(self, guid: str, formattype: str = 'YAML', export_associated=False) -> str:
         # Intended for a direct pull with no conversion
         endpoint = 'metadata/tml/export'
@@ -908,6 +1012,38 @@ class TSRestApiV1:
         else:
             raise Exception()
 
+    def metadata_tml_export_string_with_associations_map(self, guid: str, formattype: str = 'YAML') -> (str, Dict):
+        # Intended for a direct pull with no conversion
+        endpoint = 'metadata/tml/export'
+        # allow JSON or YAML in any casing
+        formattype = formattype.upper()
+        post_data = {
+            'export_ids': json.dumps([guid]),
+            'formattype': formattype,
+            'export_associated': 'true'
+        }
+        url = self.base_url + endpoint
+
+        # TML import is distinguished by having an {'Accept': 'text/plain'} header on the POST
+        response = self.session.post(url=url, data=post_data, headers={'Accept': 'text/plain'})
+        response.raise_for_status()
+        # Extra parsing of some 'error responses' that come through in JSON response on HTTP 200
+        self.raise_tml_errors(response=response)
+
+        # TML API returns a JSON response, with the TML document
+        tml_json_response = response.json()
+        objs = tml_json_response['object']
+
+        # The TML is there in full under the 'edoc' section of the API JSON response
+        response_str = objs[0]['edoc']
+
+        name_guid_map = {}
+
+        for obj in objs:
+            name_guid_map[obj['info']['name']] = obj['info']['id']
+
+        return response_str, name_guid_map
+
     # TML import is distinguished by having an {'Accept': 'text/plain'} header on the POST
     def metadata_tml_import(
         self,
@@ -1149,15 +1285,20 @@ class TSRestApiV1:
     #   used from Authenticator Server with Secret Key retrieved in a secure manner only
     #   in memory
     #
-    # def session_auth_token(self, secret_key: str, username: str, access_level: str, object_id: str):
-    #     post_params = {
-    #         'secret_key': secret_key,
-    #         'username': username,
-    #         'access_level': access_level,
-    #         'id': object_id
-    #     }
-    #    response = self.post_to_endpoint('session/auth/token', post_data=post_params)
-    #    return response
+    def session_auth_token(self, secret_key: str, username: str, access_level: str, object_id: str):
+        endpoint = 'session/auth/token'
+
+        post_params = {
+             'secret_key': secret_key,
+             'username': username,
+             'access_level': access_level,
+             'id': object_id
+        }
+
+        url = self.base_url + endpoint
+
+        response = self.session.post(url=url, data=post_params)
+        return response
 
     #
     # USER Methods
@@ -1563,6 +1704,7 @@ class TSRestApiV1:
         response.raise_for_status()
         # Returns a 204 when complete
 
+    # This is the metadata call that is used to display tables / worksheets etc. for a connection in the UI
     def connection_detail(self, connection_guid: str, sort: str = 'MODIFIED', sort_ascending: bool = True,
                           filter: Optional[str] = None, tagname: Optional[List[str]] = None, show_hidden: bool = False):
         endpoint = 'connection/detail/{}'.format(connection_guid)
@@ -1570,8 +1712,8 @@ class TSRestApiV1:
         url = self.non_public_base_url + endpoint
         url_params = {
                      'sort': sort,
-                     'sort_ascending': str(sort_ascending).lower(),
-                     'show_hidden': str(show_hidden).lower()
+                     'sortascending': str(sort_ascending).lower(),
+                     'showhidden': str(show_hidden).lower()
                      }
         if filter is not None:
             url_params['pattern'] = filter
